
RESPONSE_SUCCESS_CODES = [
    200, 
    201, 
    202, 
    204
]

VALID_REQUEST_KWARGS = [
    'method',
    'url',
    'content',
    'data',
    'files',
    'json',
    'params',
    'headers',
    'cookies',
    'auth',
    'follow_redirects',
    'timeout',
    'extensions',
]
